package no.rwr.image;

import java.awt.image.*;
import no.rwr.image.*;
import no.rwr.util.ArrayUtil;

public class BasicRendererFactory implements RendererFactory {

    static final boolean DEBUG = Renderers.DEBUG;

    public RasterOp createRenderer(ColorModel srcCM,
				   SampleModel srcSM,
				   ColorModel destCM,
				   SampleModel destSM) {

	if (srcCM.equals(destCM)) return createPixelCopier(srcSM, destSM);

	if (compatibleComponents(srcCM, destCM)) {
	    try {
		return createComponentScaler(srcCM, srcSM, destCM, destSM);
	    } catch (UnsupportedOperationException e) {
		if (DEBUG) System.err.println(e);
		// fall back/through
	    }
	}

	// Fall back on a slow color converter
	return new ColorConvertOp(srcCM, destCM);
    }

    /** @returns true if a color from one model can be converted to a
        color in the other model, simply by scaling the components of
        the color. */
    boolean compatibleComponents(ColorModel cm1, ColorModel cm2) {
	return
	    (cm1.getColorSpace().equals(cm2.getColorSpace())) &&
	    hasScalableComponents(cm1)                        &&
	    hasScalableComponents(cm2)                        &&
	    (cm1.getNumComponents() == cm2.getNumComponents());
	/* *FIXME*: is the last line redundant? Does equality on color
           spaces imply equal number of components? */
    }

    boolean hasScalableComponents(ColorModel cm) {
	return
	    (cm instanceof ComponentColorModel) ||
	    (cm instanceof PackedColorModel);
    }

    RasterOp createPixelCopier(SampleModel srcSM, SampleModel destSM) {
        int numBands = srcSM.getNumBands();
        if (destSM.getNumBands() != numBands) {
            throw new IllegalArgumentException("The number of the bands" +
                                               " of the sample models " +
                                               "do not match (" +
                                               numBands + "!=" +
                                               destSM.getNumBands() + ")");
        }
	

	if ((srcSM  instanceof ComponentSampleModel) &&
	    (destSM instanceof ComponentSampleModel)) {
            ComponentSampleModel  srcCSM = (ComponentSampleModel)  srcSM;
            ComponentSampleModel destCSM = (ComponentSampleModel) destSM;
	    if (packedSamplesMatch(srcCSM, destCSM))
		return new ComponentDataBlitOp();
	}
        return new PixelCopyOp();
    }
	
    boolean packedSamplesMatch(ComponentSampleModel  srcCSM,
			       ComponentSampleModel destCSM) {
	    
	int numBands = srcCSM.getNumBands();
	int pixelStride = srcCSM.getPixelStride();
	if (pixelStride != destCSM.getPixelStride()) return false;
	if (pixelStride > numBands) return false;

	int[]  srcOffsets =  srcCSM.getBandOffsets();
	int[] destOffsets = destCSM.getBandOffsets();
	boolean[] coverage = new boolean[numBands];

	for (int b=0; b<numBands; b++) {
	    int offset = srcOffsets[b];
	    if (offset != destOffsets[b]) return false;
	    if (offset >= numBands)       return false;
	    coverage[offset] = true;
	}

	for (int b=0; b<numBands; b++)
	    if (!coverage[b]) return false;

	return true;
    }

    RasterOp createComponentScaler(ColorModel srcCM,
				   SampleModel srcSM,
				   ColorModel destCM,
				   SampleModel destSM) {
	int numComponents = srcCM.getNumComponents();

	if (numComponents != 3) {
	    String msg = "only supports 3 components (" + numComponents +
		"!=3)";
	    throw new UnsupportedOperationException(msg);
	}

	if ((srcCM instanceof ComponentColorModel) &&
	    (destCM instanceof PackedColorModel))
	    return createComponentPacker((ComponentSampleModel) srcSM,
				  (PackedColorModel) destCM,
				  destSM);
	    
	String msg = "scaling components from " + srcCM + " to " +
	    destCM + " has not been implemented (yet)";
	
	throw new UnsupportedOperationException(msg);
    }

    RasterOp createComponentPacker(ComponentSampleModel srcCSM,
				   PackedColorModel destPCM,
				   SampleModel destSM) {	
	
	if ((srcCSM.getDataType() == DataBuffer.TYPE_BYTE)) {
	    switch (destSM.getDataType()) {
	    case DataBuffer.TYPE_USHORT:
		return new ShortFieldPacker(srcCSM, destPCM);
	    case DataBuffer.TYPE_INT:
		return new IntFieldPacker(srcCSM, destPCM);
	    }
	}
	
	String msg = "packing data type " + srcCSM.getDataType() + " into "
	    + destSM.getDataType() + " has not been implemented";
	throw new UnsupportedOperationException(msg);
    }
}
