package no.rwr.xutil;
import no.rwr.x.*;
import no.rwr.data.Buffers;
import java.awt.image.*;
import java.awt.color.*;

public class XImageFactory {

    private static final boolean DEBUG = true;

    Visual visual;
    ColorModel cm;
    SampleModel sm;

    int format;
    int bitsPerPixel;
    
    int dataType;

    int scanlineStride;


    /**
       Create a factory from an existing XImage. The main use of this
       is to create a factory that will determine the data models to
       use with this ximage. I.e., calls to getSampleModel() are
       expected. */
    public XImageFactory(XImage ximage, Visual visual) {
	this(ximage, visual, null);
    }

    public XImageFactory(XImage ximage, Visual visual, Colormap colormap) {


	this.visual = visual;
	this.format = ximage.getFormat();
	this.bitsPerPixel = ximage.getBitsPerPixel();


	// FIXME: this class currently only supports true color.

	switch (visual.getVisualClass()) {
	case Visual.VC_TRUE_COLOR:
	    initTrueColor(ximage);
	    break;

	case Visual.VC_PSEUDO_COLOR:
	    initPseudoColor(ximage, colormap);
	    break;
	    

	default:
	    if (DEBUG) System.err.println("don't know visual class");
	    throw new UnsupportedOperationException("Unsupported visual (" +
						    visual + "). " +
						    "Unable to create " +
						    "XImageFactory.");
	} 

	

    }

    void initPseudoColor(XImage ximage, Colormap colormap) {

	ColorSpace cSpace = ColorSpace.getInstance(ColorSpace.CS_sRGB);
	int dataType = Buffers.smallestAppropriateTransferType(bitsPerPixel);
	
	if (DataBuffer.getDataTypeSize(dataType) != bitsPerPixel) {
	    throw new
		UnsupportedOperationException("Unsupported number " +
					      "bits per pixel (" +
					      bitsPerPixel + "). " +
					      "No suitable data type. " +
					      "For pseudo color, the " +
					      "bits-per-pixel must " +
					      "exactly match the data " +
					      "type.");
	}
	
	XColor[] sharedColors = colormap.getSharedColors();

	int nShared = sharedColors.length;	


	byte[] rmap = new byte[nShared];
	byte[] gmap = new byte[nShared];
	byte[] bmap = new byte[nShared];
	
	for (int i=0; i < nShared; i++) {
	    rmap[i] = (byte) (sharedColors[i].getRed()   >> 8);
	    gmap[i] = (byte) (sharedColors[i].getGreen() >> 8);
	    bmap[i] = (byte) (sharedColors[i].getBlue()  >> 8);
	}

	
	cm = new IndexColorModel(8, nShared, rmap, gmap, bmap);
	
	int bytesPerLine = ximage.getBytesPerLine();		  
	this.scanlineStride = bytesPerLine * 8 / bitsPerPixel;

	int[] bandOffsets = { 0 };
	sm = new ComponentSampleModel(dataType,
				      ximage.getWidth(),
				      ximage.getHeight(),
				      1,
				      scanlineStride,
				      bandOffsets);
    }
    

    void initTrueColor(XImage ximage) {
	// True color X11 visual == Java direct color model
	//                          Java pixel packed sample model
	
	
	// I'm not sure whether the following statement is still true.
	/* FIXME: This will not work for 24 bits-per-pixel 24 bits per
	   pixel should use data type BYTE and banded sample model. */
	
	ColorSpace cSpace = ColorSpace.getInstance(ColorSpace.CS_sRGB);
	int dataType =
	    Buffers.smallestAppropriateTransferType(bitsPerPixel);
	
	if (DataBuffer.getDataTypeSize(dataType) != bitsPerPixel) {
	    throw new
		UnsupportedOperationException("Unsupported number " +
					      "bits per pixel (" +
					      bitsPerPixel + "). " +
					      "No suitable data type. " +
					      "For true color, the " +
					      "bits-per-pixel must " +
					      "exactly match the data " +
					      "type.");
	}
	
	int   redMask = ximage.  getRedMask();
	int greenMask = ximage.getGreenMask();
	int  blueMask = ximage. getBlueMask();
	
	
	// FIXME: This can be done lazily
	
	cm = new DirectColorModel(cSpace,
				  bitsPerPixel,
				  redMask,
				  greenMask,
				  blueMask,
				  0, // Alpha mask
				  false, // not alpha premultiplied
				  dataType // transfer type
				  );
	
	int bytesPerLine = ximage.getBytesPerLine();		  
	this.scanlineStride = bytesPerLine * 8 / bitsPerPixel;
	
	int[] masks = { redMask, greenMask, blueMask };
	
	sm = new SinglePixelPackedSampleModel(dataType,
					      ximage.getWidth(),
						  ximage.getHeight(),
					      scanlineStride,
					      masks);
	
	
	return;  // Done
    }
    
    

    public XImageFactory(Visual visual, ColorModel cm, SampleModel sm) {

	cm = null;
	/* An XImage does not need to be the exact same format as the
	   native xserver format. XSendImage will perform conversion.
	  
	   These attributes can be be different: (list is not exhaustive)
	   
	   bits_per_pixel
	   depth           must be one of the supported depths of the xserver

	*/

	if ((sm != null) && (cm == null)) {
	    throw new IllegalArgumentException("Cannot specify a sample " +
					       "model without specifying " +
					       "a color model. Either " +
					       "specify none, only a " +
					       "color model or both.");
	}

	this.visual = visual;
	this.cm = cm;
	this.sm = sm;


	boolean hasSubfields = visual.hasRGBSubfields();
	
	boolean isTrueColor =
	    visual.getVisualClass() == Visual.VC_TRUE_COLOR;

	// All color models in X11 have either 1 or 3 components.
	int numColorComponents = hasSubfields ? 3 : 1;

	/* FIXME: Do further test to ensure that the visual is
	   acceptable. */

	if (cm == null) {
	    /* Due to the test at the begining of the method we know
	       that (sm == null), since (cm == null). */

	    /* Without any pre-existing color- and sample models,
	       there is nothing more to do here. New models will be
	       created when needed.  */
	    if (DEBUG) { 
		System.out.println("This factory will create images that " +
				   "use the " + visual + " visual. Color- " +
				   "and sample models to match this " +
				   "visual will be created when needed.");
	    }
	    return; // All OK
	}

	/* Extract information from existing models to ensure that new
           ximages that are created by this factory are compatible
           with these models. */

	if (numColorComponents != cm.getNumColorComponents()) {
	    // FIXME: create a special ImageFormatMismatchException.
	    throw 
		new IllegalArgumentException("Visual " + visual +
					     " expects a different " +
					     "number of components than " +
					     "provided by color model " +
					     cm + " (" + numColorComponents +
					     "!=" + 
					     cm.getNumColorComponents() +
					     ")");
	}

	if (cm instanceof ComponentColorModel) {
	    ComponentColorModel ccm = (ComponentColorModel) cm;
	    ComponentSampleModel csm = (ComponentSampleModel) sm;

	    dataType = csm.getDataType();
	    int dataTypeSize = DataBuffer.getDataTypeSize(dataType);
	    int depth = visual.getDepth();


	    int[] bandOffsets = csm.getBandOffsets();
	    int maxBandOffset = 0;
	    for (int b=0; b<numColorComponents; b++) {
		maxBandOffset = Math.max(maxBandOffset, bandOffsets[b]);
	    }

	    int bitFieldSize = (maxBandOffset+1)*dataTypeSize;
	    
	    if (bitFieldSize > depth) {
		throw 
		    new IllegalArgumentException("Visual " + visual +
						 " depth (" + depth + ") " +
						 "not deep enough to " +
						 "accomodate color model " +
						 cm + " that has a " +
						 "bitField size of " +
						 bitFieldSize);
	    }

	    
	    if (numColorComponents == 3) {
		int[] masks = {
		    visual.getRedMask(),
		    visual.getGreenMask(),
		    visual.getBlueMask()
		};
		
		int[] widths = ccm.getComponentSize();
		no.rwr.util.BitMaskExtent extent = new BitMaskExtent();

		for (int c=0; c<3; c++) {
		    extent.bitWidth = (byte) widths[c];
		    extent.leastSignificantBit = (byte)
			(bandOffsets[c] * dataTypeSize);
		    if (extent.toMask() != masks[c]) {
			String msg =
			    "RGB masks of visual (" +
			    Integer.toHexString(masks[c]) + ") " +
			    "does not match the " +
			    "band distribution of the " +
			    "ComponentSampleModel (width=" + widths[c] +
			    ", offset=" + bandOffsets[c] +
			    ", dataTypeSize=" + dataTypeSize + ")";
			throw new IllegalArgumentException(msg);
		    }
		}
		
	    }

		
	    bitsPerPixel = csm.getPixelStride() * dataTypeSize;
	    scanlineStride = csm.getScanlineStride();

	    if (DEBUG) {
		System.out.println("Bit-field size: " + bitFieldSize);
		System.out.println("Visual depth: " + depth);
		System.out.println("Bits per pixel: " + bitsPerPixel);
		System.out.println("Scanline stride: " + scanlineStride);
	    }
	    




	    
	    // FIXME: don't assume
	    format = XImage.ZPIXMAP_FORMAT;
	
	    
	} else {
	    throw new UnsupportedOperationException("do not know how to " +
						    "handle color model " +
						    cm.getClass());
	}
    }

    public XImage createXImage(int width, int height) {
	// FIXME: implement

	/* FIXME: We have a problem here. The data buffer is needed in
	   the Image, but we only return an XImage. How do we avoid
	   loosing the reference to this data buffer?

	   Maybe we shouldn't create a data buffer here at all, but
	   leave this job up to the Image class? */

	return null;
    }

    public XImage createXImage(Raster raster) {

	int xoffset = 0;
	int width = raster.getWidth();
	int height = raster.getHeight();
	int bitmapPad = 32; // FIXME: don't hardcode this

	DataBuffer dataB = raster.getDataBuffer();

	if (dataType != dataB.getDataType()) {
	    throw new IllegalArgumentException("Data type mismatch between " +
					       "factory (" + dataType + ") " +
					       "and raster (" + 
					       dataB.getDataType() + ").");
	}

	int dataType = dataB.getDataType();
	int bitsPerUnit = DataBuffer.getDataTypeSize(dataType);
	int bytesPerLine = scanlineStride * bitsPerUnit / 8;
	
	XImage ximage = new XImage(visual, visual.getDepth(), format,
				   xoffset, width, height, bitmapPad,
				   bytesPerLine, bitsPerPixel);
	
	int x = raster.getMinX() - raster.getSampleModelTranslateX();
	int y = raster.getMinY() - raster.getSampleModelTranslateY();

	int offset = 
	    ((bitsPerPixel*x)/bitsPerUnit) +   // from sample model
	    scanlineStride*y;                  // from sample model
	attachData(ximage, dataB, offset);
	return ximage;
    }

    public DataBuffer createDataBuffer(XImage ximage) {
	return getSampleModel().createDataBuffer();
    }
    
    /**
       @param offset Offset to data. The given offset does not include
       data buffer offset, which will also be added.  */
    public static void
	attachData(XImage ximage, DataBuffer dataB, int offset) {
	offset += dataB.getOffset();
	switch (dataB.getDataType()) {
	case DataBuffer.TYPE_BYTE:
	    ximage.setData(((DataBufferByte) dataB).getData(), offset);
	    break;
	case DataBuffer.TYPE_USHORT:
	    ximage.setData(((DataBufferUShort) dataB).getData(), offset);
	    break;
	case DataBuffer.TYPE_INT:
	    ximage.setData(((DataBufferInt) dataB).getData(), offset);
	    break;
	default:
	    throw
		new UnsupportedOperationException("Do not know how to " +
						  "set data for data " +
						  "type " +
						  dataB.getDataType());
	}
    }
    




    public Visual getVisual() {
	return visual;
    }

    public ColorModel getColorModel() {
	// FIXME: If null, generate appropriate color model.
	return cm;
    }

    public SampleModel getSampleModel() {
	// FIXME: If null, generate appropriate sample model.
	return sm;
    }


    public String toString() {
	return getClass().getName() + "[" +
	    "visual=" + visual +
  	    ", colorModel=" + cm +
  	    ", sampleModel=" + sm +
  	    ", format=" + format +
  	    ", bitsPerPixel=" + bitsPerPixel +
  	    ", dataType=" + dataType +
  	    ", scanlineStride=" + scanlineStride + "]";
    }

}
